# -*- coding: utf-8 -*-
import os
import sys
import subprocess
import pytest

testdir = os.path.dirname(os.path.realpath(__file__))
mock_pairs_path = os.path.join(testdir, "data", "mock.4flip.pairs")
mock_chromsizes_path = os.path.join(testdir, "data", "mock.chrom.sizes")


def test_flip():
    try:
        result = subprocess.check_output(
            [
                "python",
                "-m",
                "pairtools",
                "flip",
                mock_pairs_path,
                "-c",
                mock_chromsizes_path,
            ],
        ).decode("ascii")
    except subprocess.CalledProcessError as e:
        print(e.output)
        print(sys.exc_info())
        raise e

    orig_pairs = [
        l.strip().split("\t")
        for l in open(mock_pairs_path, "r")
        if not l.startswith("#") and l.strip()
    ]
    flipped_pairs = [
        l.strip().split("\t")
        for l in result.split("\n")
        if not l.startswith("#") and l.strip()
    ]

    chrom_enum = {"!": 0, "chr1": 1, "chr2": 2, "chrU": 3, "chrU1": 4}
    # chrU stands for unannotated chromosome, which has less priority than annotated ones
    # chrU1 is another unannotated chromosome, which should go lexigographically after chrU
    for orig_pair, flipped_pair in zip(orig_pairs, flipped_pairs):
        has_correct_order = (chrom_enum[orig_pair[1]], int(orig_pair[2])) <= (
            chrom_enum[orig_pair[3]],
            int(orig_pair[4]),
        )
        if has_correct_order:
            assert all([c1 == c2 for c1, c2 in zip(orig_pair, flipped_pair)])
        if not has_correct_order:
            assert orig_pair[1] == flipped_pair[3]
            assert orig_pair[2] == flipped_pair[4]
            assert orig_pair[3] == flipped_pair[1]
            assert orig_pair[4] == flipped_pair[2]
            assert orig_pair[5] == flipped_pair[6]
            assert orig_pair[6] == flipped_pair[5]
            assert orig_pair[7] == flipped_pair[7][::-1]



import sys
import click
import warnings
import subprocess

from ..lib import fileio, pairsam_format, headerops
from ..lib.parse_pysam import AlignmentFilePairtoolized
from . import cli, common_io_options


UTIL_NAME = "pairtools_header"


@cli.group()
def header():
    """
    Manipulate the .pairs/.pairsam header
    """
    pass


# Common options for all header tools:
def register_subcommand(func):
    return header.command()(
        click.argument("pairs_path", type=str, required=False)(
            click.option(
                "-o",
                "--output",
                type=str,
                default="",
                help="output file."
                " If the path ends with .gz or .lz4, the output is bgzip-/lz4c-compressed."
                " By default, the output is printed into stdout.",
            )(
                click.option(
                    "--nproc-in",
                    type=int,
                    default=1,
                    show_default=True,
                    help="Number of processes used by the auto-guessed input decompressing command.",
                )(
                    click.option(
                        "--nproc-out",
                        type=int,
                        default=8,
                        show_default=True,
                        help="Number of processes used by the auto-guessed output compressing command.",
                    )(
                        click.option(
                            "--cmd-in",
                            type=str,
                            default=None,
                            help="A command to decompress the input. "
                            "If provided, fully overrides the auto-guessed command. "
                            "Does not work with stdin. "
                            "Must read input from stdin and print output into stdout. "
                            "EXAMPLE: pbgzip -dc -n 3",
                        )(
                            click.option(
                                "--cmd-out",
                                type=str,
                                default=None,
                                help="A command to compress the output. "
                                "If provided, fully overrides the auto-guessed command. "
                                "Does not work with stdout. "
                                "Must read input from stdin and print output into stdout. "
                                "EXAMPLE: pbgzip -c -n 8",
                            )(func)
                        )
                    )
                )
            )
        )
    )


def add_arg_help(func):
    func.__doc__ = func.__doc__.format(
        """
    PAIRS_PATH : input .pairs/.pairsam file. If the path ends with .gz or .lz4, the
    input is decompressed by bgzip/lz4c. By default, the input is read from stdin.
    """
    )
    return func


@register_subcommand
@add_arg_help
@click.option(
    "--chroms-path",
    type=str,
    default=None,
    required=False,
    help="Chromosome order used to flip interchromosomal mates: "
    "path to a chromosomes file (e.g. UCSC chrom.sizes or similar) whose "
    "first column lists scaffold names. Any scaffolds not listed will be "
    "ordered lexicographically following the names provided.",
)
@click.option(
    "--sam-path",
    type=str,
    default=None,
    required=False,
    help="Input sam file to inherit the header."
    " Either --sam or --chroms-path should be provided to store the chromosome sizes in the header.",
)
@click.option(
    "--columns",
    type=click.STRING,
    default="",
    help="Report columns describing alignments "
    "Can take multiple values as a comma-separated list."
    f"By default, assign standard .pairs columns: {','.join(pairsam_format.COLUMNS)}",
)
@click.option(
    "--extra-columns",
    type=click.STRING,
    default="",
    help="Report extra columns describing alignments "
    "Can take multiple values as a comma-separated list.",
)
@click.option(
    "--assembly",
    type=str,
    default="",
    help="Name of genome assembly (e.g. hg19, mm10) to store in the pairs header.",
)
@click.option(
    "--no-flip",
    is_flag=True,
    help="If specified, assume that the pairs are not filpped in genomic order and instead preserve "
    "the order in which they were sequenced.",
)
@click.option(
    "--pairs/--pairsam",
    is_flag=True,
    default=True,
    help=f"If pairs, then the defult columns will be set to: {','.join(pairsam_format.COLUMNS_PAIRS)}"
    f"\nif pairsam, then to: {','.join(pairsam_format.COLUMNS_PAIRSAM)}",
)
def generate(pairs_path, output, chroms_path, sam_path, columns, assembly, **kwargs):
    """
    Generate the header

    """
    generate_py(pairs_path, output, chroms_path, sam_path, columns, assembly, **kwargs)


def generate_py(pairs_path, output, chroms_path, sam_path, columns, assembly, **kwargs):

    instream = fileio.auto_open(
        pairs_path,
        mode="r",
        nproc=kwargs.get("nproc_in"),
        command=kwargs.get("cmd_in", None),
    )
    header, body_stream = headerops.get_header(instream, ignore_warning=True)

    # Parse chromosome sizes present in the input chromosomes:
    if chroms_path and not sam_path:
        chromsizes = headerops.get_chromsizes_from_file(chroms_path)
        # chromosomes = headerops.get_chromsizes_from_file(chroms_path)

    # Parse chromosome sizes present in sam input:
    if sam_path:  # open input sam file with pysam
        input_sam = AlignmentFilePairtoolized(
            sam_path, "r", threads=kwargs.get("nproc_in")
        )
        samheader = input_sam.header
        chromsizes = headerops.get_chromsizes_from_pysam_header(samheader)
        # if chroms_path:
        #     chromosomes = headerops.get_chrom_order(chroms_path, list(chromsizes.keys()))
        # else:
        #     chromosomes = chromsizes.keys()

    # Read the input columns:
    if columns:
        columns = columns.split(",")
    else:
        if kwargs.get("pairs", True):
            columns = pairsam_format.COLUMNS_PAIRS
        else:
            columns = pairsam_format.COLUMNS_PAIRSAM

    extra_columns = kwargs.get("extra_columns", "")
    if extra_columns:
        columns += extra_columns.split(",")

    # Write new header to the pairsam file
    new_header = headerops.make_standard_pairsheader(
        assembly=assembly,
        chromsizes=chromsizes,
        columns=columns,
        shape="whole matrix" if kwargs["no_flip"] else "upper triangle",
    )

    if sam_path:
        new_header = headerops.insert_samheader_pysam(new_header, samheader)

    new_header = headerops.append_new_pg(new_header, ID=UTIL_NAME, PN=UTIL_NAME)

    # Check that the number of columns in the body corresponds to the header:
    if not headerops.validate_cols(instream, columns):
        raise ValueError(
            f"Number of columns mismatch:\n\t#columns: {headerops.SEP_COLS.join(columns)}\n\t{body_stream.readline()}"
        )

    ########
    # Write the output after successful checks:
    outstream = (
        fileio.auto_open(
            output,
            mode="w",
            nproc=kwargs.get("nproc_out"),
            command=kwargs.get("cmd_out", None),
        )
        if output
        else sys.stdout
    )

    outstream.writelines((l + "\n" for l in new_header))
    outstream.flush()

    if body_stream == sys.stdin:
        for line in body_stream:
            outstream.write(line)
    else:
        command = r"""
                   /bin/bash -c 'export LC_COLLATE=C; export LANG=C; cat """

        if kwargs.get("cmd_in", None):
            command += r""" <(cat {} | {} | sed -n -e '\''/^[^#]/,$p'\'')""".format(
                pairs_path, kwargs["cmd_in"]
            )
        elif pairs_path.endswith(".gz"):
            command += (
                r""" <(bgzip -dc -@ {} {} | sed -n -e '\''/^[^#]/,$p'\'')""".format(
                    kwargs["nproc_in"], pairs_path
                )
            )
        elif pairs_path.endswith(".lz4"):
            command += r""" <(lz4c -dc {} | sed -n -e '\''/^[^#]/,$p'\'')""".format(
                pairs_path
            )
        else:
            command += r""" <(sed -n -e '\''/^[^#]/,$p'\'' {})""".format(pairs_path)
        command += "'"

        subprocess.check_call(command, shell=True, stdout=outstream)

    if instream != sys.stdin:
        instream.close()

    if outstream != sys.stdout:
        outstream.close()


@register_subcommand
@add_arg_help
@click.option(
    "--reference-file", "-r", help="Header file for transfer", type=str, required=True
)
def transfer(pairs_path, output, reference_file, **kwargs):
    """
    Transfer the header from one pairs file to another

    """
    transfer_py(pairs_path, output, reference_file, **kwargs)


def transfer_py(pairs_path, output, reference_file, **kwargs):

    instream = fileio.auto_open(
        pairs_path,
        mode="r",
        nproc=kwargs.get("nproc_in"),
        command=kwargs.get("cmd_in", None),
    )
    header, body_stream = headerops.get_header(instream, ignore_warning=True)

    # Read the header from reference file
    instream_header = fileio.auto_open(
        reference_file,
        mode="r",
        nproc=kwargs.get("nproc_in"),
        command=kwargs.get("cmd_in", None),
    )
    reference_header, _ = headerops.get_header(instream_header)
    # Close the reference stream after extraction of the header:
    if instream_header != sys.stdin:
        instream_header.close()

    reference_columns = headerops.extract_column_names(reference_header)

    # Check that the number of columns in the body corresponds to the header:
    if not headerops.validate_cols(instream, reference_columns):
        raise ValueError(
            f"Number of columns mismatch:\n\t#columns: {headerops.SEP_COLS.join(reference_columns)}\n\t{body_stream.readline()}"
        )

    ########
    # Write the output after successful checks:
    outstream = (
        fileio.auto_open(
            output,
            mode="w",
            nproc=kwargs.get("nproc_out"),
            command=kwargs.get("cmd_out", None),
        )
        if output
        else sys.stdout
    )

    reference_header = headerops.append_new_pg(
        reference_header, ID=UTIL_NAME, PN=UTIL_NAME
    )
    outstream.writelines((l + "\n" for l in reference_header))
    outstream.flush()

    if body_stream == sys.stdin:
        for line in body_stream:
            outstream.write(line)
    else:
        command = r"""
                   /bin/bash -c 'export LC_COLLATE=C; export LANG=C; cat """

        if kwargs.get("cmd_in", None):
            command += r""" <(cat {} | {} | sed -n -e '\''/^[^#]/,$p'\'')""".format(
                pairs_path, kwargs["cmd_in"]
            )
        elif pairs_path.endswith(".gz"):
            command += (
                r""" <(bgzip -dc -@ {} {} | sed -n -e '\''/^[^#]/,$p'\'')""".format(
                    kwargs["nproc_in"], pairs_path
                )
            )
        elif pairs_path.endswith(".lz4"):
            command += r""" <(lz4c -dc {} | sed -n -e '\''/^[^#]/,$p'\'')""".format(
                pairs_path
            )
        else:
            command += r""" <(sed -n -e '\''/^[^#]/,$p'\'' {})""".format(pairs_path)
        command += "'"

        subprocess.check_call(command, shell=True, stdout=outstream)

    if instream != sys.stdin:
        instream.close()

    if outstream != sys.stdout:
        outstream.close()


@register_subcommand
@add_arg_help
@click.option(
    "--columns",
    "-c",
    help=f"Comma-separated list of columns to be added, e.g.: {','.join(pairsam_format.COLUMNS)}",
    type=str,
    required=True,
)
def set_columns(pairs_path, output, columns, **kwargs):
    """
    Add the columns to the .pairs/pairsam file
    """
    set_columns_py(pairs_path, output, columns, **kwargs)


def set_columns_py(pairs_path, output, columns, **kwargs):
    instream = (
        fileio.auto_open(
            pairs_path,
            mode="r",
            nproc=kwargs.get("nproc_in"),
            command=kwargs.get("cmd_in", None),
        )
        if pairs_path
        else sys.stdin
    )
    outstream = (
        fileio.auto_open(
            output,
            mode="w",
            nproc=kwargs.get("nproc_out"),
            command=kwargs.get("cmd_out", None),
        )
        if output
        else sys.stdout
    )

    header, body_stream = headerops.get_header(instream)
    header = headerops.set_columns(header, columns.split(","))
    outstream.writelines((l + "\n" for l in header))
    outstream.flush()

    if body_stream == sys.stdin:
        for line in body_stream:
            outstream.write(line)
    else:
        command = r"""
                   /bin/bash -c 'export LC_COLLATE=C; export LANG=C; cat """

        if kwargs.get("cmd_in", None):
            command += r""" <(cat {} | {} | sed -n -e '\''/^[^#]/,$p'\'')""".format(
                pairs_path, kwargs["cmd_in"]
            )
        elif pairs_path.endswith(".gz"):
            command += (
                r""" <(bgzip -dc -@ {} {} | sed -n -e '\''/^[^#]/,$p'\'')""".format(
                    kwargs["nproc_in"], pairs_path
                )
            )
        elif pairs_path.endswith(".lz4"):
            command += r""" <(lz4c -dc {} | sed -n -e '\''/^[^#]/,$p'\'')""".format(
                pairs_path
            )
        else:
            command += r""" <(sed -n -e '\''/^[^#]/,$p'\'' {})""".format(pairs_path)
        command += "'"

        subprocess.check_call(command, shell=True, stdout=outstream)

    if instream != sys.stdin:
        instream.close()

    if outstream != sys.stdout:
        outstream.close()


@register_subcommand
@add_arg_help
@click.option(
    "--reference-file",
    "-r",
    help="Header file for comparison (optional)",
    type=str,
    required=False,
    default="",
)
@click.option(
    "--reference-columns",
    "-c",
    help=f"Comma-separated list of columns fro check (optional), e.g.: {','.join(pairsam_format.COLUMNS)}",
    type=str,
    required=False,
    default="",
)
def validate_columns(pairs_path, output, reference_file, reference_columns, **kwargs):
    """
    Validate the columns of the .pairs/pairsam file against reference or within file.
    If the checks pass, then returns full pairs file. Otherwise throws an exception.

    If reference_file is provided, check:
        1) columns are the same between pairs and reference_file
        2) number of columns in the pairs body is the same as the number of columns

    If reference_columns are provided, check:
        1) pairs columns are the same as provided
        2) number of columns in the pairs body is the same as the number of columns

    If no reference_file or columns, then check only the number of columns in the pairs body.
    Checks only the first line in the pairs stream!

    """
    validate_columns_py(pairs_path, output, reference_file, reference_columns, **kwargs)


def validate_columns_py(
    pairs_path, output, reference_file, reference_columns, **kwargs
):

    instream = fileio.auto_open(
        pairs_path,
        mode="r",
        nproc=kwargs.get("nproc_in"),
        command=kwargs.get("cmd_in", None),
    )
    header, body_stream = headerops.get_header(instream)
    pairs_columns = headerops.extract_column_names(header)

    # Convert reference columns string into list, if provided
    if reference_columns:
        reference_columns = reference_columns.split(",")

    # Read the header from reference file
    if reference_file:
        instream_header = fileio.auto_open(
            reference_file,
            mode="r",
            nproc=kwargs.get("nproc_in"),
            command=kwargs.get("cmd_in", None),
        )
        reference_header, _ = headerops.get_header(instream_header)
        # Close the reference stream after extraction of the header:
        if instream_header != sys.stdin:
            instream_header.close()

        if reference_columns:
            warnings.warn(
                "--reference-columns are ignored, as --reference-file is provided"
            )

        reference_columns = headerops.extract_column_names(reference_header)

    if reference_columns:
        if pairs_columns != reference_columns:
            raise ValueError(
                f"Pairs columns differ from reference columns:\n\t{pairs_columns}\n\t{reference_columns}"
            )

    # Check that the number of columns in the body corresponds to the header:
    if not headerops.validate_cols(instream, pairs_columns):
        raise ValueError(
            f"Number of columns mismatch:\n\t#columns: {headerops.SEP_COLS.join(pairs_columns)}\n\t{body_stream.readline()}"
        )

    ########
    # Write the output after successful checks:
    outstream = (
        fileio.auto_open(
            output,
            mode="w",
            nproc=kwargs.get("nproc_out"),
            command=kwargs.get("cmd_out", None),
        )
        if output
        else sys.stdout
    )
    header = headerops.append_new_pg(header, ID=UTIL_NAME, PN=UTIL_NAME)
    outstream.writelines((l + "\n" for l in header))
    outstream.flush()

    if body_stream == sys.stdin:
        for line in body_stream:
            outstream.write(line)
    else:
        command = r"""
                   /bin/bash -c 'export LC_COLLATE=C; export LANG=C; cat """

        if kwargs.get("cmd_in", None):
            command += r""" <(cat {} | {} | sed -n -e '\''/^[^#]/,$p'\'')""".format(
                pairs_path, kwargs["cmd_in"]
            )
        elif pairs_path.endswith(".gz"):
            command += (
                r""" <(bgzip -dc -@ {} {} | sed -n -e '\''/^[^#]/,$p'\'')""".format(
                    kwargs["nproc_in"], pairs_path
                )
            )
        elif pairs_path.endswith(".lz4"):
            command += r""" <(lz4c -dc {} | sed -n -e '\''/^[^#]/,$p'\'')""".format(
                pairs_path
            )
        else:
            command += r""" <(sed -n -e '\''/^[^#]/,$p'\'' {})""".format(pairs_path)
        command += "'"

        subprocess.check_call(command, shell=True, stdout=outstream)

    if instream != sys.stdin:
        instream.close()

    if outstream != sys.stdout:
        outstream.close()


if __name__ == "__main__":
    header()



# -*- coding: utf-8 -*-

import click
import functools
import sys
from .. import __version__
import logging
from .._logging import get_logger


CONTEXT_SETTINGS = {
    "help_option_names": ["-h", "--help"],
}


@click.version_option(version=__version__)
@click.group(context_settings=CONTEXT_SETTINGS)
@click.option(
    "--post-mortem", help="Post mortem debugging", is_flag=True, default=False
)
@click.option(
    "--output-profile",
    help="Profile performance with Python cProfile and dump the statistics "
    "into a binary file",
    type=str,
    default="",
)
@click.option("-v", "--verbose", help="Verbose logging.", count=True)
@click.option(
    "-d",
    "--debug",
    help="On error, drop into the post-mortem debugger shell.",
    is_flag=True,
    default=False,
)
def cli(post_mortem, output_profile, verbose, debug):
    """Flexible tools for Hi-C data processing.

    All pairtools have a few common options, which should be typed _before_
    the command name.

    """
    if post_mortem:
        import traceback

        try:
            import ipdb as pdb
        except ImportError:
            import pdb

        def _excepthook(exc_type, value, tb):
            traceback.print_exception(exc_type, value, tb)
            print()
            pdb.pm()

        sys.excepthook = _excepthook

    if output_profile:
        import cProfile
        import atexit

        pr = cProfile.Profile()
        pr.enable()

        def _atexit_profile_hook():
            pr.disable()
            pr.dump_stats(output_profile)

        atexit.register(_atexit_profile_hook)

    # Initialize logging to stderr
    logging.basicConfig(stream=sys.stderr)
    logging.captureWarnings(True)
    root_logger = get_logger()

    # Set verbosity level
    if verbose > 0:
        root_logger.setLevel(logging.DEBUG)
        if verbose > 1:  # pragma: no cover
            try:
                import psutil
                import atexit

                @atexit.register
                def process_dump_at_exit():
                    process_attrs = [
                        "cmdline",
                        # 'connections',
                        "cpu_affinity",
                        "cpu_num",
                        "cpu_percent",
                        "cpu_times",
                        "create_time",
                        "cwd",
                        # 'environ',
                        "exe",
                        # 'gids',
                        "io_counters",
                        "ionice",
                        "memory_full_info",
                        # 'memory_info',
                        # 'memory_maps',
                        "memory_percent",
                        "name",
                        "nice",
                        "num_ctx_switches",
                        "num_fds",
                        "num_threads",
                        "open_files",
                        "pid",
                        "ppid",
                        "status",
                        "terminal",
                        "threads",
                        # 'uids',
                        "username",
                    ]
                    p = psutil.Process()
                    info_ = p.as_dict(process_attrs, ad_value="")
                    for key in process_attrs:
                        root_logger.debug("PSINFO:'{}': {}".format(key, info_[key]))

            except ImportError:
                root_logger.warning("Install psutil to see process information.")

    else:
        root_logger.setLevel(logging.INFO)

    # Set hook for postmortem debugging
    if debug:  # pragma: no cover
        import traceback

        try:
            import ipdb as pdb
        except ImportError:
            import pdb

        def _excepthook(exc_type, value, tb):
            traceback.print_exception(exc_type, value, tb)
            print()
            pdb.pm()

        sys.excepthook = _excepthook


def common_io_options(func):
    @click.option(
        "--nproc-in",
        type=int,
        default=3,
        show_default=True,
        help="Number of processes used by the auto-guessed input decompressing command.",
    )
    @click.option(
        "--nproc-out",
        type=int,
        default=8,
        show_default=True,
        help="Number of processes used by the auto-guessed output compressing command.",
    )
    @click.option(
        "--cmd-in",
        type=str,
        default=None,
        help="A command to decompress the input file. "
        "If provided, fully overrides the auto-guessed command. "
        "Does not work with stdin and pairtools parse. "
        "Must read input from stdin and print output into stdout. "
        "EXAMPLE: pbgzip -dc -n 3",
    )
    @click.option(
        "--cmd-out",
        type=str,
        default=None,
        help="A command to compress the output file. "
        "If provided, fully overrides the auto-guessed command. "
        "Does not work with stdout. "
        "Must read input from stdin and print output into stdout. "
        "EXAMPLE: pbgzip -c -n 8",
    )
    @functools.wraps(func)
    def wrapper(*args, **kwargs):
        return func(*args, **kwargs)

    return wrapper


from . import (
    dedup,
    sort,
    flip,
    merge,
    markasdup,
    select,
    split,
    restrict,
    phase,
    parse,
    parse2,
    stats,
    sample,
    filterbycov,
    header,
    scaling,
)



# -*- coding: utf-8 -*-
import os
import sys
import subprocess
import pytest
import tempfile

testdir = os.path.dirname(os.path.realpath(__file__))

mock_pairs_path_filterbycov = os.path.join(testdir, "data", "mock.4filterbycov.pairs")

tmpdir = tempfile.TemporaryDirectory()
tmpdir_name = tmpdir.name


params = [
    {"max_dist": 0, "max_cov": 3},
    {"max_dist": 0, "max_cov": 2},
    {"max_dist": 1, "max_cov": 1},
]

for p in params:
    p["lowcov_path"] = os.path.join(
        tmpdir_name, "lowcov.{}.{}.pairs".format(p["max_dist"], p["max_cov"])
    )
    p["highcov_path"] = os.path.join(
        tmpdir_name, "highcov.{}.{}.pairs".format(p["max_dist"], p["max_cov"])
    )
    p["unmapped_path"] = os.path.join(
        tmpdir_name, "unmapped.{}.{}.pairs".format(p["max_dist"], p["max_cov"])
    )


@pytest.fixture
def setup_filterbycov():
    try:
        for p in params:
            subprocess.check_output(
                [
                    "python",
                    "-m",
                    "pairtools",
                    "filterbycov",
                    mock_pairs_path_filterbycov,
                    "--output",
                    p["lowcov_path"],
                    "--output-highcov",
                    p["highcov_path"],
                    "--output-unmapped",
                    p["unmapped_path"],
                    "--max-dist",
                    str(p["max_dist"]),
                    "--max-cov",
                    str(p["max_cov"]),
                ]
            )
    except subprocess.CalledProcessError as e:
        print(e.output)
        print(sys.exc_info())
        raise e


def test_mock_pairs(setup_filterbycov):

    all_pairs = [
        l.strip().split("\t")
        for l in open(mock_pairs_path_filterbycov, "r")
        if not l.startswith("#") and l.strip()
    ]
    for p in params:

        lowcov_pairs = [
            l.strip().split("\t")
            for l in open(p["lowcov_path"], "r")
            if not l.startswith("#") and l.strip()
        ]
        highcov_pairs = [
            l.strip().split("\t")
            for l in open(p["highcov_path"], "r")
            if not l.startswith("#") and l.strip()
        ]
        unmapped_pairs = [
            l.strip().split("\t")
            for l in open(p["unmapped_path"], "r")
            if not l.startswith("#") and l.strip()
        ]

        # check that at least a few pairs remained in deduped and dup files
        # assert len(lowcov_pairs) > 0
        # assert len(highcov_pairs) > 0
        # assert len(unmapped_pairs) > 0

        # check that all pairs entries survived deduping:

        assert len(lowcov_pairs) + len(unmapped_pairs) + len(highcov_pairs) == len(
            all_pairs
        )

        assert all([(pair[1] != "!" and pair[3] != "!") for pair in lowcov_pairs])
        assert all([(pair[1] != "!" and pair[3] != "!") for pair in highcov_pairs])
        assert all([(pair[1] == "!" or pair[3] == "!") for pair in unmapped_pairs])

        def update_coverage(coverage, chrom, pos, max_dist):
            if chrom == "!":
                return

            coverage[chrom] = coverage.get(chrom, {})
            for i in range(max(0, pos - max_dist), pos + max_dist + 1):
                coverage[chrom][i] = coverage[chrom].get(i, 0) + 1

        coverage = {}
        for pair in all_pairs:
            update_coverage(coverage, pair[1], int(pair[2]), p["max_dist"])
            update_coverage(coverage, pair[3], int(pair[4]), p["max_dist"])

        for pair in lowcov_pairs:
            # print (p['max_cov'],p['max_dist'])
            # print (pair, coverage[pair[1]][int(pair[2])])
            # print (pair, coverage[pair[3]][int(pair[4])])
            assert coverage[pair[1]][int(pair[2])] <= p["max_cov"]
            assert coverage[pair[3]][int(pair[4])] <= p["max_cov"]

        for pair in highcov_pairs:
            # print (p['max_cov'],p['max_dist'])
            # print (pair, coverage[pair[1]][int(pair[2])])
            # print (pair, coverage[pair[3]][int(pair[4])])
            assert (coverage[pair[1]][int(pair[2])] > p["max_cov"]) or (
                coverage[pair[3]][int(pair[4])] > p["max_cov"]
            )

    tmpdir.cleanup()



# -*- coding: utf-8 -*-
import os
import sys
import subprocess
import pytest

testdir = os.path.dirname(os.path.realpath(__file__))


def test_mock_pairsam():
    mock_pairsam_path = os.path.join(testdir, "data", "mock.pairsam")
    try:
        result = subprocess.check_output(
            ["python", "-m", "pairtools", "sort", mock_pairsam_path],
        ).decode("ascii")
    except subprocess.CalledProcessError as e:
        print(e.output)
        print(sys.exc_info())
        raise e

    # Check that the only changes strings are a @PG record of a SAM header,
    # the "#sorted" entry and chromosomes
    pairsam_header = [
        l.strip() for l in open(mock_pairsam_path, "r") if l.startswith("#")
    ]
    output_header = [l.strip() for l in result.split("\n") if l.startswith("#")]

    print(output_header)
    print(pairsam_header)
    for l in output_header:
        if not any([l in l2 for l2 in pairsam_header]):
            assert (
                l.startswith("#samheader: @PG")
                or l.startswith("#sorted")
                or l.startswith("#chromosomes")
            )

    pairsam_body = [
        l.strip()
        for l in open(mock_pairsam_path, "r")
        if not l.startswith("#") and l.strip()
    ]
    output_body = [
        l.strip() for l in result.split("\n") if not l.startswith("#") and l.strip()
    ]

    # check that all pairsam entries survived sorting:
    assert len(pairsam_body) == len(output_body)

    # check the sorting order of the output:
    prev_pair = None
    for l in output_body:
        cur_pair = l.split("\t")[1:8]
        if prev_pair is not None:
            assert cur_pair[0] >= prev_pair[0]
            if cur_pair[0] == prev_pair[0]:
                assert cur_pair[2] >= prev_pair[2]
                if cur_pair[2] == prev_pair[2]:
                    assert cur_pair[1] >= prev_pair[1]
                    if cur_pair[1] == prev_pair[1]:
                        assert cur_pair[3] >= prev_pair[3]

        prev_pair = cur_pair



#!/usr/bin/env python
# -*- coding: utf-8 -*-
import io
import os
import re
import glob


from setuptools import find_packages, setup
from setuptools.extension import Extension

try:
    from Cython.Distutils import build_ext as _build_ext
    from Cython.Build import cythonize

    HAVE_CYTHON = True
except ImportError:
    from setuptools.command.build_ext import build_ext as _build_ext

    HAVE_CYTHON = False

classifiers = """\
    Development Status :: 4 - Beta
    Operating System :: OS Independent
    Programming Language :: Python
    Programming Language :: Python :: 3
    Programming Language :: Python :: 3.7
    Programming Language :: Python :: 3.8
    Programming Language :: Python :: 3.9
    Programming Language :: Python :: 3.10
"""


def _read(*parts, **kwargs):
    filepath = os.path.join(os.path.dirname(__file__), *parts)
    encoding = kwargs.pop("encoding", "utf-8")
    with io.open(filepath, encoding=encoding) as fh:
        text = fh.read()
    return text


def get_version():
    version = re.search(
        r'^__version__\s*=\s*[\'"]([^\'"]*)[\'"]',
        _read("pairtools", "__init__.py"),
        re.MULTILINE,
    ).group(1)
    return version


long_description = _read("README.md")

install_requires = [l for l in _read("requirements.txt").split("\n") if l]


def get_ext_modules():
    ext = ".pyx" if HAVE_CYTHON else ".c"
    src_files = glob.glob(
        os.path.join(os.path.dirname(__file__), "pairtools", "lib", "*" + ext)
    )

    ext_modules = []
    for src_file in src_files:
        name = "pairtools.lib." + os.path.splitext(os.path.basename(src_file))[0]
        if not "pysam" in name and not "regions" in name:
            ext_modules.append(Extension(name, [src_file]))
        elif "regions" in name:
            ext_modules.append(
                Extension(
                    name,
                    [src_file],
                    language="c++",
                )
            )
        else:
            import pysam
            ext_modules.append(
                Extension(
                    name,
                    [src_file],
                    extra_link_args=pysam.get_libraries(),
                    include_dirs=pysam.get_include(),
                    define_macros=pysam.get_defines(),
                    #extra_objects=pysam.get_libraries(),
                )
            )

    if HAVE_CYTHON:
        # .pyx to .c
        ext_modules = cythonize(ext_modules)  # , annotate=True

    return ext_modules


class build_ext(_build_ext):
    # Extension module build configuration
    def finalize_options(self):
        _build_ext.finalize_options(self)
        # Fix to work with bootstrapped numpy installation
        # http://stackoverflow.com/a/21621689/579416
        # Prevent numpy from thinking it is still in its setup process:
        __builtins__.__NUMPY_SETUP__ = False
        import numpy

        self.include_dirs.append(numpy.get_include())

    def run(self):

        # Import numpy here, only when headers are needed
        import numpy

        # Add numpy headers to include_dirs
        self.include_dirs.append(numpy.get_include())

        # Call original build_ext command
        _build_ext.run(self)


setup(
    name="pairtools",
    author="Open2C",
    author_email="open.chromosome.collective@gmail.com",
    version=get_version(),
    license="MIT",
    description="CLI tools to process mapped Hi-C data",
    long_description=long_description,
    long_description_content_type="text/markdown",
    keywords=["genomics", "bioinformatics", "Hi-C", "contact"],
    url="https://github.com/open2c/pairtools",
    ext_modules=get_ext_modules(),
    cmdclass={"build_ext": build_ext},
    zip_safe=False,
    classifiers=[s.strip() for s in classifiers.split("\n") if s],
    install_requires=install_requires,
    python_requires=">=3.7",
    entry_points={
        "console_scripts": [
            "pairtools = pairtools.cli:cli",
            #'pairsamtools = pairtools.cli:cli',
        ]
    },
    packages=find_packages(),
)


#!/usr/bin/env python
# -*- coding: utf-8 -*-
import sys
import click
import warnings

import numpy as np

from ..lib import fileio, pairsam_format, headerops
from . import cli, common_io_options

from ..lib.restrict import find_rfrag

UTIL_NAME = "pairtools_restrict"


@cli.command()
@click.argument("pairs_path", type=str, required=False)
@click.option(
    "-f",
    "--frags",
    type=str,
    required=True,
    help="a tab-separated BED file with the positions of restriction fragments "
    "(chrom, start, end). Can be generated using cooler digest.",
)
@click.option(
    "-o",
    "--output",
    type=str,
    default="",
    help="output .pairs/.pairsam file."
    " If the path ends with .gz/.lz4, the output is compressed by bgzip/lz4c."
    " By default, the output is printed into stdout.",
)
@common_io_options
def restrict(pairs_path, frags, output, **kwargs):
    """Assign restriction fragments to pairs.

    Identify the restriction fragments that got ligated into a Hi-C molecule.

    Note: rfrags are 0-indexed

    PAIRS_PATH : input .pairs/.pairsam file. If the path ends with .gz/.lz4, the
    input is decompressed by bgzip/lz4c. By default, the input is read from stdin.
    """
    restrict_py(pairs_path, frags, output, **kwargs)


def restrict_py(pairs_path, frags, output, **kwargs):
    instream = fileio.auto_open(
        pairs_path,
        mode="r",
        nproc=kwargs.get("nproc_in"),
        command=kwargs.get("cmd_in", None),
    )

    outstream = fileio.auto_open(
        output,
        mode="w",
        nproc=kwargs.get("nproc_out"),
        command=kwargs.get("cmd_out", None),
    )

    header, body_stream = headerops.get_header(instream)
    header = headerops.append_new_pg(header, ID=UTIL_NAME, PN=UTIL_NAME)
    header = headerops.append_columns(
        header,
        [
            "rfrag1",
            "rfrag_start1",
            "rfrag_end1",
            "rfrag2",
            "rfrag_start2",
            "rfrag_end2",
        ],
    )

    outstream.writelines((l + "\n" for l in header))
    rfrags = np.genfromtxt(
        frags,
        delimiter="\t",
        comments="#",
        dtype=None,
        encoding="ascii",
        names=["chrom", "start", "end"],
    )

    rfrags.sort(order=["chrom", "start", "end"])
    chrom_borders = np.r_[
        0, 1 + np.where(rfrags["chrom"][:-1] != rfrags["chrom"][1:])[0], rfrags.shape[0]
    ]
    rfrags = {
        rfrags["chrom"][i]: np.concatenate([[0], rfrags["end"][i:j] + 1])
        for i, j in zip(chrom_borders[:-1], chrom_borders[1:])
    }

    for line in body_stream:
        cols = line.rstrip('\n').split(pairsam_format.PAIRSAM_SEP)
        chrom1, pos1 = cols[pairsam_format.COL_C1], int(cols[pairsam_format.COL_P1])
        rfrag_idx1, rfrag_start1, rfrag_end1 = find_rfrag(rfrags, chrom1, pos1)
        chrom2, pos2 = cols[pairsam_format.COL_C2], int(cols[pairsam_format.COL_P2])
        rfrag_idx2, rfrag_start2, rfrag_end2 = find_rfrag(rfrags, chrom2, pos2)
        cols += [str(rfrag_idx1), str(rfrag_start1), str(rfrag_end1)]
        cols += [str(rfrag_idx2), str(rfrag_start2), str(rfrag_end2)]
        outstream.write(pairsam_format.PAIRSAM_SEP.join(cols))
        outstream.write("\n")

    if instream != sys.stdin:
        instream.close()
    if outstream != sys.stdout:
        outstream.close()


if __name__ == "__main__":
    restrict()



#!/usr/bin/env python
import sys
import glob
import math
import subprocess
import click

from ..lib import fileio, pairsam_format, headerops
from . import cli, common_io_options

UTIL_NAME = "pairtools_merge"


@cli.command()
@click.argument(
    "pairs_path",
    nargs=-1,
    type=str,
)
@click.option(
    "-o",
    "--output",
    type=str,
    default="",
    help="output file."
    " If the path ends with .gz/.lz4, the output is compressed by bgzip/lz4c."
    " By default, the output is printed into stdout.",
)
@click.option(
    "--max-nmerge",
    type=int,
    default=8,
    show_default=True,
    help="The maximal number of inputs merged at once. For more, store "
    "merged intermediates in temporary files.",
)
@click.option(
    "--tmpdir",
    type=str,
    default="",
    help="Custom temporary folder for merged intermediates.",
)
@click.option(
    "--memory",
    type=str,
    default="2G",
    show_default=True,
    help="The amount of memory used by default.",
)
@click.option(
    "--compress-program",
    type=str,
    default="",
    show_default=True,
    help="A binary to compress temporary merged chunks. "
    "Must decompress input when the flag -d is provided. "
    "Suggested alternatives: lz4c, gzip, lzop, snzip. "
    "NOTE: fails silently if the command syntax is wrong. ",
)
@click.option(
    "--nproc",
    type=int,
    default=8,
    help="Number of threads for merging.",
    show_default=True,
)
@click.option(
    "--nproc-in",
    type=int,
    default=1,
    show_default=True,
    help="Number of processes used by the auto-guessed input decompressing command.",
)
@click.option(
    "--nproc-out",
    type=int,
    default=8,
    show_default=True,
    help="Number of processes used by the auto-guessed output compressing command.",
)
@click.option(
    "--cmd-in",
    type=str,
    default=None,
    help="A command to decompress the input. "
    "If provided, fully overrides the auto-guessed command. "
    "Does not work with stdin. "
    "Must read input from stdin and print output into stdout. "
    "EXAMPLE: pbgzip -dc -n 3",
)
@click.option(
    "--cmd-out",
    type=str,
    default=None,
    help="A command to compress the output. "
    "If provided, fully overrides the auto-guessed command. "
    "Does not work with stdout. "
    "Must read input from stdin and print output into stdout. "
    "EXAMPLE: pbgzip -c -n 8",
)
@click.option(
    "--keep-first-header/--no-keep-first-header",
    default=False,
    show_default=True,
    help="Keep the first header or merge the headers together. Default: merge headers.",
)
@click.option(
    "--concatenate/--no-concatenate",
    default=False,
    show_default=True,
    help="Simple concatenate instead of merging sorted files.",
)
# Using custom IO options


def merge(
    pairs_path, output, max_nmerge, tmpdir, memory, compress_program, nproc, **kwargs
):
    """Merge .pairs/.pairsam files.
    By default, assumes that the files are sorted and maintains the sorting.

    Merge triu-flipped sorted pairs/pairsam files. If present, the @SQ records
    of the SAM header must be identical; the sorting order of
    these lines is taken from the first file in the list.
    The ID fields of the @PG records of the SAM header are modified with a
    numeric suffix to produce unique records.
    The other unique SAM and non-SAM header lines are copied into the output header.

    PAIRS_PATH : upper-triangular flipped sorted .pairs/.pairsam files to merge
    or a group/groups of .pairs/.pairsam files specified by a wildcard. For
    paths ending in .gz/.lz4, the files are decompressed by bgzip/lz4c.

    """
    merge_py(
        pairs_path,
        output,
        max_nmerge,
        tmpdir,
        memory,
        compress_program,
        nproc,
        **kwargs,
    )


def merge_py(
    pairs_path, output, max_nmerge, tmpdir, memory, compress_program, nproc, **kwargs
):
    paths = sum([glob.glob(mask) for mask in pairs_path], [])

    if len(paths) == 0:
        raise ValueError(f"No input paths: {pairs_path}")

    outstream = fileio.auto_open(
        output,
        mode="w",
        nproc=kwargs.get("nproc_out"),
        command=kwargs.get("cmd_out", None),
    )

    # if there is only one input, bypass merging and do not modify the header
    if len(paths) == 1:
        instream = fileio.auto_open(
            paths[0],
            mode="r",
            nproc=kwargs.get("nproc_in"),
            command=kwargs.get("cmd_in", None),
        )
        for line in instream:
            outstream.write(line)
        if outstream != sys.stdout:
            outstream.close()

        return

    headers = []
    for path in paths:
        f = fileio.auto_open(
            path,
            mode="r",
            nproc=kwargs.get("nproc_in"),
            command=kwargs.get("cmd_in", None),
        )
        h, _ = headerops.get_header(f)
        headers.append(h)
        f.close()
        # Skip other headers if keep_first_header is True (False by default):
        if kwargs.get("keep_first_header", False):
            break

    if not headerops.all_same_columns(headers):
        raise ValueError("Input pairs cannot contain different columns")

    merged_header = headerops.merge_headers(headers)
    merged_header = headerops.append_new_pg(merged_header, ID=UTIL_NAME, PN=UTIL_NAME)

    outstream.writelines((l + "\n" for l in merged_header))
    outstream.flush()

    # If concatenation requested instead of merging sorted input:
    if kwargs.get("concatenate", False):
        command = r"""
                    /bin/bash -c 'export LC_COLLATE=C; export LANG=C; cat """
    # Full merge that keeps the ordered input:
    else:
        command = r"""
            /bin/bash -c 'export LC_COLLATE=C; export LANG=C; sort
            -k {0},{0} -k {1},{1} -k {2},{2}n -k {3},{3}n -k {4},{4} 
            --merge  
            --field-separator=$'\''{5}'\''
            {6}
            {7}
            {8}
            -S {9}
            {10}
            """.replace(
            "\n", " "
        ).format(
            pairsam_format.COL_C1 + 1,
            pairsam_format.COL_C2 + 1,
            pairsam_format.COL_P1 + 1,
            pairsam_format.COL_P2 + 1,
            pairsam_format.COL_PTYPE + 1,
            pairsam_format.PAIRSAM_SEP_ESCAPE,
            " --parallel={} ".format(nproc) if nproc > 1 else " ",
            " --batch-size={} ".format(max_nmerge) if max_nmerge else " ",
            " --temporary-directory={} ".format(tmpdir) if tmpdir else " ",
            memory,
            (
                " --compress-program={} ".format(compress_program)
                if compress_program
                else " "
            ),
        )
    for path in paths:
        if kwargs.get("cmd_in", None):
            command += r""" <(cat {} | {} | sed -n -e '\''/^[^#]/,$p'\'')""".format(
                path, kwargs["cmd_in"]
            )
        elif path.endswith(".gz"):
            command += (
                r""" <(bgzip -dc -@ {} {} | sed -n -e '\''/^[^#]/,$p'\'')""".format(
                    kwargs["nproc_in"], path
                )
            )
        elif path.endswith(".lz4"):
            command += r""" <(lz4c -dc {} | sed -n -e '\''/^[^#]/,$p'\'')""".format(
                path
            )
        else:
            command += r""" <(sed -n -e '\''/^[^#]/,$p'\'' {})""".format(path)
    command += "'"

    subprocess.check_call(command, shell=True, stdout=outstream)

    if outstream != sys.stdout:
        outstream.close()


if __name__ == "__main__":
    merge()



# -*- coding: utf-8 -*-
import os
import sys
import subprocess
import pytest

testdir = os.path.dirname(os.path.realpath(__file__))


def test_mock_pairsam():
    mock_pairsam_path = os.path.join(testdir, "data", "mock.pairsam")
    try:
        result = subprocess.check_output(
            ["python", "-m", "pairtools", "markasdup", mock_pairsam_path],
        ).decode("ascii")
    except subprocess.CalledProcessError as e:
        print(e.output)
        print(sys.exc_info())
        raise e

    pairsam_body = [
        l.strip()
        for l in open(mock_pairsam_path, "r")
        if not l.startswith("#") and l.strip()
    ]
    output_body = [
        l.strip() for l in result.split("\n") if not l.startswith("#") and l.strip()
    ]

    # check that all pairsam entries survived sorting:
    assert len(pairsam_body) == len(output_body)

    # check that all pairtypes got changed to DD
    for l in output_body:
        assert l.split("\t")[7] == "DD"



def get_chrom_phase(chrom, phase_suffixes):
    if chrom.endswith(phase_suffixes[0]):
        return "0", chrom[: -len(phase_suffixes[0])]
    elif chrom.endswith(phase_suffixes[1]):
        return "1", chrom[: -len(phase_suffixes[1])]
    else:
        return "!", chrom


def phase_side_XB(chrom, XB, AS, XS, phase_suffixes):

    phase, chrom_base = get_chrom_phase(chrom, phase_suffixes)

    XBs = [i for i in XB.split(";") if len(i) > 0]
    S1, S2, S3 = AS, XS, -1  # -1 if the second hit was not reported

    if AS > XS:  # Primary hit has higher score than the secondary
        return phase, chrom_base, S1, S2, S3

    elif len(XBs) >= 1:
        if len(XBs) >= 2:
            alt2_chrom, alt2_pos, alt2_CIGAR, alt2_NM, alt2_AS, alt_mapq = XBs[1].split(
                ","
            )
            S3 = int(alt2_AS)
            if int(alt2_AS) == XS == AS:
                return "!", "!", S1, S2, S3

        alt_chrom, alt_pos, alt_CIGAR, alt_NM, alt_AS, alt_mapq = XBs[0].split(",")
        alt_phase, alt_chrom_base = get_chrom_phase(alt_chrom, phase_suffixes)

        alt_is_homologue = (chrom_base == alt_chrom_base) and (
            ((phase == "0") and (alt_phase == "1"))
            or ((phase == "1") and (alt_phase == "0"))
        )

        if alt_is_homologue:
            return ".", chrom_base, S1, S2, S3

    return "!", "!", S1, S2, S3


def phase_side_XA(chrom, XA, AS, XS, NM, phase_suffixes):

    phase, chrom_base = get_chrom_phase(chrom, phase_suffixes)

    XAs = [i for i in XA.split(";") if len(i.strip()) > 0]
    if len(XAs) >= 1:
        alt_chrom, alt_pos, alt_CIGAR, alt_NM = XAs[0].split(",")
        M1, M2, M3 = NM, int(alt_NM), -1
    else:
        M1, M2, M3 = NM, -1, -1  # -1 if the second hit was not reported

    if AS > XS:  # Primary hit has higher score than the secondary
        return phase, chrom_base, M1, M2, M3

    elif len(XAs) >= 1:

        if len(XAs) >= 2:
            alt2_chrom, alt2_pos, alt2_CIGAR, alt2_NM = XAs[1].split(",")
            M3 = int(alt2_NM)
            if int(alt2_NM) == int(alt_NM) == NM:
                return "!", "!", M1, M2, M3

        alt_chrom, alt_pos, alt_CIGAR, alt_NM = XAs[0].split(",")

        alt_phase, alt_chrom_base = get_chrom_phase(alt_chrom, phase_suffixes)

        alt_is_homologue = (chrom_base == alt_chrom_base) and (
            ((phase == "0") and (alt_phase == "1"))
            or ((phase == "1") and (alt_phase == "0"))
        )

        if alt_is_homologue:
            return ".", chrom_base, M1, M2, M3

    return "!", "!", M1, M2, M3



from ..lib import fileio, pairsam_format, headerops
import re, fnmatch

# Create environment of important functions:
wildcard_library = {}


def wildcard_match(x, wildcard):
    if wildcard not in wildcard_library:
        regex = fnmatch.translate(wildcard)
        reobj = re.compile(regex)
        wildcard_library[wildcard] = reobj
    return wildcard_library[wildcard].fullmatch(x)


csv_library = {}


def csv_match(x, csv):
    if csv not in csv_library:
        csv_library[csv] = set(csv.split(","))
    return x in csv_library[csv]


regex_library = {}


def regex_match(x, regex):
    if regex not in regex_library:
        reobj = re.compile(regex)
        regex_library[regex] = reobj
    return regex_library[regex].fullmatch(x)


# Define default data types:
TYPES = {"pos1": "int", "pos2": "int", "mapq1": "int", "mapq2": "int"}


def evaluate_stream(
    headerless_stream, condition, column_names, type_cast=(), startup_code=None
):
    """
    Evaluate expression for the input headerless stream.

    Parameters
    ----------
    headerless_stream
    condition
    type_cast: Cast a given column to a given type. By default, only pos and mapq
                are cast to int, other columns are kept as str. Type: tupe of two strings.
    startup_code: An auxiliary code to execute before filtering.
                Use to define functions that can be evaluated in the CONDITION statement

    ========
    Writes the output to one of two streams (regular or rest)

    """

    # Define data types:
    TYPES.update(dict(type_cast))

    # Execute startup code:
    if startup_code is not None:
        exec(startup_code, globals())

    for i, col in enumerate(column_names):
        if col in TYPES:
            col_type = TYPES[col]
            condition = re.sub(r"\b%s\b" % col , "{}(COLS[{}])".format(col_type, i), condition)
            #condition.replace(col, "{}(COLS[{}])".format(col_type, i))
        else:
            condition = re.sub(r"\b%s\b" % col, "COLS[{}]".format(i), condition)
            #condition = condition.replace(col, "COLS[{}]".format(i))

    # Compile the filtering expression:
    match_func = compile(condition, "<string>", "eval")

    for line in headerless_stream:
        COLS = line.rstrip().split(pairsam_format.PAIRSAM_SEP)

        # Evaluate filtering expression:
        filter_passed = eval(match_func)

        # Produce the output:
        yield filter_passed, line


def evaluate_df(df, condition, type_cast=(), startup_code=None, engine="pandas"):
    """
    Evaluate expression for the input headerless stream.

    Parameters
    ----------
    df: input dataframe for evaluation
    condition: condition to evaluate
    type_cast: additional types transformations, if different from default
    startup_code: An auxiliary code to execute before filtering.
                Use to define functions that can be evaluated in the CONDITION statement

    ========
    Writes the output to one of two streams (regular or rest)

    """

    # Define data types:
    TYPES.update(dict(type_cast))

    # Execute startup code:
    if startup_code is not None:
        exec(startup_code, globals())

    # Set up the column formats:
    for col in df.columns:
        if col in TYPES.keys():
            if not str(df.dtypes[col]) != TYPES[col]:
                df[col] = df[col].astype(TYPES[col])

    if engine == "pandas":
        try:
            filter_passed_output = df.eval(condition)
        except ValueError as e:
            raise ValueError(f"Try passing engine python to fix this: {e}")
    else:
        # Set up the columns indexing
        for i, col in enumerate(df.columns):
            condition = re.sub(r"\b%s\b" % col, "COLS[{}]".format(i), condition)
            #condition = condition.replace(col, "COLS[{}]".format(i))

        filter_passed_output = []
        match_func = compile(condition, "<string>", "eval")
        for i, r in df.iterrows():
            COLS = r.values

            # Evaluate filtering expression:
            filter_passed = eval(match_func)
            filter_passed_output.append(True if filter_passed else False)

    return filter_passed_output



""" Moved from pairlib, library for fast regions assignment """
from cython.operator cimport dereference, postincrement, postdecrement

from cpython cimport array
import cython

from libcpp.map cimport map
from libcpp.algorithm cimport lower_bound, upper_bound
from libcpp.string cimport string
from libcpp.vector cimport vector

import numpy as np
cimport numpy as np

cpdef np.ndarray assign_regs_c(np.ndarray chroms, np.ndarray pos, dict reg_dict):
    assert len(chroms) == len(pos)

    cdef int n = len(chroms)
    cdef np.ndarray[np.int64_t, ndim=2] result = -1 * np.ones((n, 3), dtype=np.int64)
    cdef map[string, vector[int]] reg_map = reg_dict

    cdef map[string, vector[int]].iterator reg_map_it = reg_map.begin()
    cdef map[string, vector[int]].iterator reg_map_end = reg_map.end()

    cdef vector[int].iterator lo_b, up_b
    cdef int position, reg_boundary_idx

    # this can be parallelized with prange
    for i in range(n):
        reg_map_it = reg_map.find(chroms[i])
        if reg_map_it != reg_map_end:
            position = pos[i]
            up_b = upper_bound(
                dereference(reg_map_it).second.begin(),
                dereference(reg_map_it).second.end(),
                position)
            reg_boundary_idx = up_b - dereference(reg_map_it).second.begin()

            if reg_boundary_idx % 2 == 1:
                lo_b = up_b
                postdecrement(lo_b)
                result[i, 0] = (reg_boundary_idx - 1) // 2
                result[i, 1] = dereference(lo_b)
                result[i, 2] = dereference(up_b)

    return result


#!/usr/bin/env python
# -*- coding: utf-8 -*-
import io
import sys
import click

import pandas as pd
from ..lib import fileio, pairsam_format, headerops
from . import cli, common_io_options

from ..lib.stats import PairCounter, do_merge

from .._logging import get_logger

logger = get_logger()

UTIL_NAME = "pairtools_stats"


@cli.command()
@click.argument("input_path", type=str, nargs=-1, required=False)
@click.option("-o", "--output", type=str, default="", help="output stats tsv file.")
@click.option(
    "--merge",
    is_flag=True,
    help="If specified, merge multiple input stats files instead of calculating"
    " statistics of a .pairs/.pairsam file. Merging is performed via summation of"
    " all overlapping statistics. Non-overlapping statistics are appended to"
    " the end of the file. Supported for tsv stats with single filter.",
)
@click.option(
    "--with-chromsizes/--no-chromsizes",
    is_flag=True,
    default=True,
    help="If enabled, will store sizes of chromosomes from the header of the pairs file"
    " in the stats file.",
)
@click.option(
    "--yaml/--no-yaml",
    is_flag=True,
    default=False,
    help="Output stats in yaml format instead of table. ",
)
@click.option(
    "--bytile-dups/--no-bytile-dups",
    default=False,
    help="If enabled, will analyse by-tile duplication statistics to estimate"
    " library complexity more accurately."
    " Requires parent_readID column to be saved by dedup (will be ignored otherwise)"
    " Saves by-tile stats into --output_bytile-stats stream, or regular output if --output_bytile-stats is not provided.",
)
@click.option(
    "--output-bytile-stats",
    default="",
    required=False,
    help="output file for tile duplicate statistics."
    " If file exists, it will be open in the append mode."
    " If the path ends with .gz or .lz4, the output is bgzip-/lz4c-compressed."
    " By default, by-tile duplicate statistics are not printed."
    " Note that the readID and parent_readID should be provided and contain tile information for this option.",
)
# Filtering options:
@click.option(
    "--filter",
    default=None,
    required=False,
    multiple=True,
    help="Filters with conditions to apply to the data (similar to `pairtools select`). "
    "For non-YAML output only the first filter will be reported. "
    """Example: pairtools stats --yaml --filter 'unique:(pair_type=="UU")' --filter 'close:(pair_type=="UU") and (abs(pos1-pos2)<10)' test.pairs """,
)
@click.option(
    "--engine",
    default="pandas",
    required=False,
    help="Engine for regular expression parsing. "
    "Python will provide you regex functionality, while pandas does not accept custom funtctions and works faster. ",
)
@click.option(
    "--chrom-subset",
    type=str,
    default=None,
    required=False,
    help="A path to a chromosomes file (tab-separated, 1st column contains "
    "chromosome names) containing a chromosome subset of interest. "
    "If provided, additionally filter pairs with both sides originating from "
    "the provided subset of chromosomes. This operation modifies the #chromosomes: "
    "and #chromsize: header fields accordingly.",
)
@click.option(
    "--startup-code",
    type=str,
    default=None,
    required=False,
    help="An auxiliary code to execute before filtering. "
    "Use to define functions that can be evaluated in the CONDITION statement",
)
@click.option(
    "-t",
    "--type-cast",
    type=(str, str),
    default=(),
    multiple=True,
    help="Cast a given column to a given type. By default, only pos and mapq "
    "are cast to int, other columns are kept as str. Provide as "
    "-t <column_name> <type>, e.g. -t read_len1 int. Multiple entries are allowed.",
)
@common_io_options
def stats(
    input_path, output, merge, bytile_dups, output_bytile_stats, filter, **kwargs
):
    """Calculate pairs statistics.

    INPUT_PATH : by default, a .pairs/.pairsam file to calculate statistics.
    If not provided, the input is read from stdin.
    If --merge is specified, then INPUT_PATH is interpreted as an arbitrary number
    of stats files to merge.

    The files with paths ending with .gz/.lz4 are decompressed by bgzip/lz4c.
    """

    stats_py(
        input_path,
        output,
        merge,
        bytile_dups,
        output_bytile_stats,
        filter,
        **kwargs,
    )


def stats_py(
    input_path, output, merge, bytile_dups, output_bytile_stats, filter, **kwargs
):
    if merge:
        do_merge(output, input_path, **kwargs)
        return

    if len(input_path) == 0:
        raise ValueError(f"No input paths: {input_path}")

    instream = fileio.auto_open(
        input_path[0],
        mode="r",
        nproc=kwargs.get("nproc_in"),
        command=kwargs.get("cmd_in", None),
    )
    outstream = fileio.auto_open(
        output,
        mode="w",
        nproc=kwargs.get("nproc_out"),
        command=kwargs.get("cmd_out", None),
    )
    if bytile_dups and not output_bytile_stats:
        output_bytile_stats = outstream
    if output_bytile_stats:
        bytile_dups = True

    header, body_stream = headerops.get_header(instream)
    cols = headerops.extract_column_names(header)

    # Check necessary columns for reporting by-tile stats:
    if bytile_dups and "parent_readID" not in cols:
        logger.warning(
            "No 'parent_readID' column in the file, not generating duplicate stats."
        )
        bytile_dups = False

    # Define filters and their properties
    first_filter_name = "no_filter"  # default filter name for full output
    if filter is not None and len(filter) > 0:
        first_filter_name = filter[0].split(":", 1)[0]
        if len(filter) > 1 and not kwargs.get("yaml", False):
            logger.warn(
                f"Output the first filter only in non-YAML output: {first_filter_name}"
            )

        filter = dict([f.split(":", 1) for f in filter])
    else:
        filter = None

    stats = PairCounter(
        bytile_dups=bytile_dups,
        filters=filter,
        startup_code=kwargs.get("startup_code", ""),  # for evaluation of filters
        type_cast=kwargs.get("type_cast", ()),  # for evaluation of filters
        engine=kwargs.get("engine", "pandas"),
    )

    # Collecting statistics
    for chunk in pd.read_table(body_stream, names=cols, chunksize=100_000):
        stats.add_pairs_from_dataframe(chunk)

    if kwargs.get("with_chromsizes", True):
        chromsizes = headerops.extract_chromsizes(header)
        stats.add_chromsizes(chromsizes)

    if bytile_dups:
        stats.save_bytile_dups(output_bytile_stats)

    # save statistics to file ...
    stats.save(
        outstream,
        yaml=kwargs.get("yaml", False),  # format as yaml
        filter=first_filter_name
        if not kwargs.get("yaml", False)
        else None,  # output only the first filter if non-YAML output
    )

    if instream != sys.stdin:
        instream.close()
    if outstream != sys.stdout:
        outstream.close()


if __name__ == "__main__":
    stats()
